import pandas as pd


def get_number_of_containing_entries(freq_1, freq_2):
    """
    Calculate the number of entries of a date range object that
    are contained in another date range object.

    :param freq_1: str of a Pandas offset alias
    :param freq_2: str of a Pandas offset alias
    :return: Integer
    """
    # Create date range objects
    start_date = "2000-01-01"
    end_date = "2010-01-01"
    range_1 = pd.date_range(start=start_date, end=end_date, freq=freq_1)
    range_2 = pd.date_range(start=start_date, end=end_date, freq=freq_2)

    # Calculate number of containing entries
    if len(range_1) < len(range_2):
        res = len(range_2) / len(range_1)
    else:
        res = len(range_1) / len(range_2)

    return round(res)



from abc import ABCMeta

from ontime.abstract.abstract_base_model import AbstractBaseModel
from ontime.time_series import TimeSeries

from skforecast.ForecasterAutoreg import (
    ForecasterAutoreg as SkForecastForecasterAutoreg,
)


class ForecasterAutoreg(AbstractBaseModel):
    """
    Generic wrapper around SkForecast ForecasterAutoreg models
    """

    def __init__(self, model, **params):
        super().__init__()
        if isinstance(model, ABCMeta):
            model = model()
        self.model = SkForecastForecasterAutoreg(regressor=model, **params)

    def fit(self, ts, **params):
        self.model.fit(y=ts.pd_series(), **params)
        return self

    def predict(self, n, **params):
        pred = self.model.predict(n, **params)
        return TimeSeries.from_series(pred)



from ontime.abstract.abstract_base_model import AbstractBaseModel
from ontime.time_series import TimeSeries


class ForecastingModel(AbstractBaseModel):
    """
    Generic wrapper around Darts forecasting models
    """

    def __init__(self, model, **params):
        super().__init__()
        self.model = model(**params)

    def fit(self, ts, **params):
        self.model.fit(ts, **params)
        return self

    def predict(self, n, **params):
        pred = self.model.predict(n, **params)
        return TimeSeries.from_darts(pred)



from darts.models import CatBoostModel
from darts.utils.statistics import check_seasonality
import ontime as on


class GenericPredictor:
    """
    Generic predictor for univariate time series using CatBoost model
    """

    def __init__(self):
        self.model = None

    def fit(self, ts):
        """
        Fit the model to the given time series
        :param ts: TimeSeries
        :return: self
        """
        # Get information about the time series
        has_seasonality, seasonality = check_seasonality(ts)
        lags = 12 if seasonality == 0 else seasonality

        # Create model
        self.model = on.Model(
            CatBoostModel,
            lags=int(lags),
        )
        self.model.fit(ts)
        return self

    def predict(self, n):
        """
        Predict n steps into the future
        :param n: Int number of steps to predict
        :return: TimeSeries
        """
        return self.model.predict(n)



from abc import ABC, abstractmethod
from typing import NoReturn


class AbstractBaseGenerator(ABC):
    """Abstract class to define methods to implement
    for a Generator class.
    """

    @abstractmethod
    def generate(self, **kwargs) -> NoReturn:
        """Generate features"""
        raise NotImplementedError



# Generic
DEFAULT_EXPORT_FILENAME = "export"

# TimeSeries
TIME_SERIES_FILENAME = "data"
TIME_SERIES_EXT = "csv"

# Component
COMPONENT_VALUES = "values"
COMPONENT_META_PREFIX = "meta_"
COMPONENT_META_LABEL_PREFIX = "label_"
COMPONENT_META_CI_LOWER = "ci_lower"
COMPONENT_META_CI_UPPER = "ci_upper"

# Metadata
METADATA_FILENAME = "meta"
METADATA_EXT = "json"
METADATA_CLASS_LABEL = "label"

# Model
MODEL_TYPE_UNIVARIATE = "univariate"
MODEL_TYPE_MULTIVARIATE = "multivariate"

# Misc. IO
PICKLE_EXT = "pkl"



from darts.models.forecasting.forecasting_model import ModelMeta

from ontime.abstract.abstract_base_model import AbstractBaseModel
from ontime.time_series import TimeSeries

from .libs.darts.forecasting_model import ForecastingModel as DartsForecastingModel
from .libs.skforecast.forecaster_autoreg import (
    ForecasterAutoreg as SkForecastForecasterAutoreg,
)


class Model(AbstractBaseModel):
    """
    Generic wrapper around all implemented time series libraries
    """

    def __init__(self, model, **params):
        super().__init__()
        if isinstance(model, ModelMeta):
            # Darts Models
            self.model = DartsForecastingModel(model, **params)
        else:
            # scikit-learn API compatible models
            self.model = SkForecastForecasterAutoreg(model, **params)

    def fit(self, ts, **params):
        """
        Fit the model to the given time series
        :param ts: TimeSeries
        :param params: Parameters to pass to the model
        :return: self
        """
        self.model.fit(ts, **params)
        return self

    def predict(self, n, **params):
        """
        Predict n steps into the future
        :param n: Integer
        :param params: Parameters to pass to the predict method
        :return: TimeSeries
        """
        pred = self.model.predict(n, **params)
        return TimeSeries.from_darts(pred)



from darts.dataprocessing.transformers.missing_values_filler import (
    MissingValuesFiller as DartsMissingValuesFiller,
)

from ...abstract import AbstractBaseProcessor
from ...time_series import TimeSeries


class Filler(AbstractBaseProcessor):
    """Wrapper around Darts MissingValuesFiller.
    https://unit8co.github.io/darts/generated_api/darts.dataprocessing.transformers.missing_values_filler.html
    """

    def __init__(self, fill="auto", name="Fill", n_jobs=1, verbose=False):
        """Data transformer to fill missing values from a (sequence of) deterministic ``TimeSeries``.

        :param fill: Union[str, Union[int, float]]
            The value used to replace the missing values.
            If set to 'auto', will auto-fill missing values using the :func:`pd.Dataframe.interpolate()` method.
        :param name: str
            A specific name for the transformer
        :param n_jobs: int
            The number of jobs to run in parallel. Parallel jobs are created only when a ``Sequence[TimeSeries]`` is
            passed as input to a method, parallelising operations regarding different ``TimeSeries``. Defaults to `1`
            (sequential). Setting the parameter to `-1` means using all the available processors.
            Note: for a small amount of data, the parallelisation overhead could end up increasing the total
            required amount of time.
        :param verbose: bool
            Optionally, whether to print operations progress
        """
        super().__init__()
        self._processor = DartsMissingValuesFiller(
            fill=fill, name=name, n_jobs=n_jobs, verbose=verbose
        )

    def process(self, ts) -> TimeSeries:
        """Process the time series.
        :param ts: TimeSeries
        :return: TimeSeries
        """
        ts_processed = self._processor.transform(ts)
        return TimeSeries.from_darts(ts_processed)



class Registry:
    def __init__(self):
        self.registry = {}

    def register(self, name, item):
        self.registry[name] = item

    def get(self, name):
        return self.registry[name]

    def get_all(self):
        return list(self.registry.keys())

    def update(self, name, item):
        self.registry[name] = item

    def delete(self, name):
        del self.registry[name]



from typing import Sequence, Union

from darts.ad.detectors.quantile_detector import QuantileDetector

from ...abstract import AbstractBaseDetector
from ...time_series import TimeSeries


class Quantile(QuantileDetector, AbstractBaseDetector):
    """
    Wrapper around Darts QuantileDetector.
    """

    def __init__(self, low_quantile=None, high_quantile=None):
        super().__init__(low_quantile, high_quantile)

    def fit(self, ts: TimeSeries) -> None:
        """
        Fits the detector to the given time series.
        :param ts: TimeSeries
        :return: None
        """
        super().fit(ts)

    def detect(self, ts: TimeSeries) -> TimeSeries:
        """
        Detects anomalies in the given time series.
        :param ts: TimeSeries
        :return: TimeSeries
        """
        ts_detected = super().detect(ts)
        return TimeSeries.from_darts(ts_detected)



import pandas as pd
import altair as alt


def line(ts):
    """
    Standard line plot for TimeSeries
    :param ts: TimeSeries
    :return: Altair Chart
    """
    alt.data_transformers.enable("vegafusion")

    # Transform data
    df = ts.pd_dataframe()
    df = df.reset_index()
    df = df.melt("time", var_name="variable", value_name="value")

    # Plot
    chart = (
        alt.Chart(df)
        .mark_line()
        .encode(
            x="time:T",
            y="value:Q",
            color="variable:N",
        )
        .properties(width=500, height=300)
    )

    return chart


def heatmap(ts):
    """
    Plot a Heatmap of a TimeSeries
    :param ts: TimeSeries
    :return: Altair Chart
    """
    alt.data_transformers.enable("vegafusion")

    # Transform data
    df = ts.pd_dataframe()
    df = df.reset_index()
    df = df.melt("time", var_name="variable", value_name="value")

    color_condition = alt.condition(
        "month(datum.value) == 1 && date(datum.value) == 1",
        alt.value("black"),
        alt.value(None),
    )

    chart = (
        alt.Chart(df)
        .mark_rect()
        .encode(
            alt.X("yearmonthdate(time):O")
            .title("Time")
            .axis(
                format="%Y",
                labelAngle=0,
                labelOverlap=False,
                labelColor=color_condition,
                tickColor=color_condition,
            ),
            alt.Y("variable:N").title(None),
            alt.Color("sum(value)").title("Value"),
        )
        .properties(width=500, height=100)
    )

    return chart


def prediction(train_ts, pred_ts=None, test_ts=None):
    """
    Plot a prediction
    :param train_ts: TimeSeries
    :param pred_ts: TimeSeries
    :param test_ts: TimeSeries
    :return: Altair Chart
    """
    alt.data_transformers.enable("vegafusion")

    # Train section
    df_train = train_ts.pd_dataframe()
    df_train = df_train.reset_index()
    df_train = df_train.melt("time", var_name="variable", value_name="value")
    df_train["variable"] = "Training set"

    # Prediction section
    df_pred = pred_ts.pd_dataframe()
    df_pred = df_pred.reset_index()
    df_pred = df_pred.melt("time", var_name="variable", value_name="value")
    df_pred["variable"] = "Prediction"

    # Test section
    df_test = test_ts.pd_dataframe()
    df_test = df_test.reset_index()
    df_test = df_test.melt("time", var_name="variable", value_name="value")
    df_test["variable"] = "Truth"

    df = pd.concat([df_train, df_pred, df_test])

    chart = (
        alt.Chart(df)
        .mark_line()
        .encode(
            x="time:T",
            y="value:Q",
            color="variable:N",
            opacity=alt.condition(
                alt.datum.variable == "Truth", alt.value(0.5), alt.value(1.0)
            ),
            strokeDash=alt.condition(
                alt.datum.variable == "Truth",
                alt.value(
                    [5, 2]
                ),  # Dash pattern: 5 units of line followed by 2 units of gap
                alt.value([0]),  # Solid line
            ),
        )
        .properties(width=500, height=300)
    )

    return chart


def anomalies(ts, ts_anomaly):
    """
    Plot Anomalies
    :param ts: normal series
    :param ts_anomaly: anomaly series
    :return: Altair Chart
    """
    alt.data_transformers.enable("vegafusion")

    df = ts.pd_dataframe()
    df = df.reset_index()
    df = df.melt("time", var_name="variable", value_name="value")
    df["variable"] = "signal"

    df_anomaly = ts_anomaly.pd_dataframe()
    df_anomaly = df_anomaly.reset_index()
    df_anomaly = df_anomaly.melt("time", var_name="variable", value_name="value")
    df_anomaly["variable"] = "anomaly"

    df = pd.concat([df, df_anomaly])
    df = df.pivot(index="time", columns="variable", values="value").reset_index()

    chart_signal = (
        alt.Chart(df)
        .mark_line()
        .encode(x="time:T", y="signal:Q")
        .properties(width=500, height=300)
    )

    chart_anomaly = (
        alt.Chart(df)
        .transform_filter(alt.datum.anomaly == 1)
        .mark_circle(color="red", size=60)
        .encode(
            x="time:T",
            y="signal:Q",
        )
    )

    return chart_signal + chart_anomaly



from typing import Optional, Tuple, Union, List

from darts.models.forecasting.catboost_model import CatBoostModel

from ...abstract import AbstractBaseModel
from ...time_series import TimeSeries


class CatBoost(CatBoostModel, AbstractBaseModel):
    """
    Wrapper around Darts CatBoost model.
    """

    def __init__(
        self,
        lags: Union[int, list] = None,
        lags_past_covariates: Union[int, List[int]] = None,
        lags_future_covariates: Union[Tuple[int, int], List[int]] = None,
        output_chunk_length: int = 1,
        add_encoders: Optional[dict] = None,
        likelihood: str = None,
        quantiles: List = None,
        random_state: Optional[int] = None,
        multi_models: Optional[bool] = True,
        use_static_covariates: bool = True,
        **kwargs,
    ):
        """
        CatBoost Model

        :param lags:
             Lagged target values used to predict the next time step. If an integer is given the last `lags` past lags
            are used (from -1 backward). Otherwise a list of integers with lags is required (each lag must be < 0).
        :param lags_past_covariates:
            Number of lagged past_covariates values used to predict the next time step. If an integer is given the last
            `lags_past_covariates` past lags are used (inclusive, starting from lag -1). Otherwise a list of integers
            with lags < 0 is required.
        :param lags_future_covariates:
            Number of lagged future_covariates values used to predict the next time step. If an tuple (past, future) is
            given the last `past` lags in the past are used (inclusive, starting from lag -1) along with the first
            `future` future lags (starting from 0 - the prediction time - up to `future - 1` included). Otherwise a list
            of integers with lags is required.
        :param output_chunk_length:
            Number of time steps predicted at once by the internal regression model. Does not have to equal the forecast
            horizon `n` used in `predict()`. However, setting `output_chunk_length` equal to the forecast horizon may
            be useful if the covariates don't extend far enough into the future.
        :param add_encoders:
            A large number of past and future covariates can be automatically generated with `add_encoders`.
            This can be done by adding multiple pre-defined index encoders and/or custom user-made functions that
            will be used as index encoders. Additionally, a transformer such as Darts' :class:`Scaler` can be added to
            transform the generated covariates. This happens all under one hood and only needs to be specified at
            model creation.
            Read :meth:`SequentialEncoder <darts.dataprocessing.encoders.SequentialEncoder>` to find out more about
            ``add_encoders``. Default: ``None``. An example showing some of ``add_encoders`` features:
                add_encoders={
                    'cyclic': {'future': ['month']},
                    'datetime_attribute': {'future': ['hour', 'dayofweek']},
                    'position': {'past': ['relative'], 'future': ['relative']},
                    'custom': {'past': [lambda idx: (idx.year - 1950) / 50]},
                    'transformer': Scaler()
                }
        :param likelihood:
            Can be set to 'quantile', 'poisson' or 'gaussian'. If set, the model will be probabilistic,
            allowing sampling at prediction time. When set to 'gaussian', the model will use CatBoost's
            'RMSEWithUncertainty' loss function. When using this loss function, CatBoost returns a mean
            and variance couple, which capture data (aleatoric) uncertainty.
            This will overwrite any `objective` parameter.
        :param quantiles:
            Fit the model to these quantiles if the `likelihood` is set to `quantile`.
        :param random_state:
            Control the randomness in the fitting procedure and for sampling.
            Default: ``None``.
        :param multi_models:
            If True, a separate model will be trained for each future lag to predict. If False, a single model is
            trained to predict at step 'output_chunk_length' in the future. Default: True.
        :param use_static_covariates:
            Whether the model should use static covariate information in case the input `series` passed to ``fit()``
            contain static covariates. If ``True``, and static covariates are available at fitting time, will enforce
            that all target `series` have the same static covariate dimensionality in ``fit()`` and ``predict()``.
        :param kwargs:
            Additional keyword arguments passed to `catboost.CatBoostRegressor`.
        """
        super().__init__(
            lags,
            lags_past_covariates,
            lags_future_covariates,
            output_chunk_length,
            add_encoders,
            likelihood,
            quantiles,
            random_state,
            multi_models,
            use_static_covariates,
            **kwargs,
        )

    def fit(
        self,
        ts: TimeSeries,
        past_covariates: TimeSeries = None,
        future_covariates: Optional[TimeSeries] = None,
        val_series: Optional[TimeSeries] = None,
        val_past_covariates: Optional[TimeSeries] = None,
        val_future_covariates: Optional[TimeSeries] = None,
        max_samples_per_ts: Optional[int] = None,
        verbose: Optional[Union[int, bool]] = 0,
        **kwargs,
    ):
        """Fits/trains the model using the provided list of features time series and the target time series.

        :param ts:
            TimeSeries or Sequence[TimeSeries] object containing the target values.
        :param past_covariates:
            Optionally, a series or sequence of series specifying past-observed covariates
        :param future_covariates:
            Optionally, a series or sequence of series specifying future-known covariates
        :param val_series:
            TimeSeries or Sequence[TimeSeries] object containing the target values for evaluation dataset
        :param val_past_covariates:
            Optionally, a series or sequence of series specifying past-observed covariates for evaluation dataset
        :param val_future_covariates : Union[TimeSeries, Sequence[TimeSeries]]
            Optionally, a series or sequence of series specifying future-known covariates for evaluation dataset
        :param max_samples_per_ts:
            This is an integer upper bound on the number of tuples that can be produced
            per time series. It can be used in order to have an upper bound on the total size of the dataset and
            ensure proper sampling. If `None`, it will read all of the individual time series in advance (at dataset
            creation) to know their sizes, which might be expensive on big datasets.
            If some series turn out to have a length that would allow more than `max_samples_per_ts`, only the
            most recent `max_samples_per_ts` samples will be considered.
        :param verbose:
            An integer or a boolean that can be set to 1 to display catboost's default verbose output
        :param **kwargs:
            Additional kwargs passed to `catboost.CatboostRegressor.fit()`
        :return: self
        """
        super().fit(
            ts,
            past_covariates,
            future_covariates,
            val_series,
            val_past_covariates,
            val_future_covariates,
            max_samples_per_ts,
            verbose,
            **kwargs,
        )
        return self

    def predict(
        self,
        n: int,
        ts: Optional[TimeSeries] = None,
        past_covariates: Optional[TimeSeries] = None,
        future_covariates: Optional[TimeSeries] = None,
        num_samples: int = 1,
        verbose: bool = False,
        predict_likelihood_parameters: bool = False,
        **kwargs,
    ) -> TimeSeries:
        """Forecasts values for `n` time steps after the end of the series.

        :param n: int
            Forecast horizon - the number of time steps after the end of the series for which to produce predictions.
        :param ts: TimeSeries, optional
            Optionally, one or several input `TimeSeries`, representing the history of the target series whose future
            is to be predicted. If specified, the method returns the forecasts of these series. Otherwise, the method
            returns the forecast of the (single) training series.
        :param past_covariates: TimeSeries or list of TimeSeries, optional
            Optionally, the past-observed covariates series needed as inputs for the model.
            They must match the covariates used for training in terms of dimension and type.
        :param future_covariates: TimeSeries or list of TimeSeries, optional
            Optionally, the future-known covariates series needed as inputs for the model.
            They must match the covariates used for training in terms of dimension and type.
        :param num_samples: int, default: 1
            Number of times a prediction is sampled from a probabilistic model. Should be set to 1
            for deterministic models.
        :param verbose:
            Optionally, whether to print progress.
        :param predict_likelihood_parameters:
            If set to `True`, the model predict the parameters of its Likelihood parameters instead of the target. Only
            supported for probabilistic models with a likelihood, `num_samples = 1` and `n<=output_chunk_length`.
            Default: ``False``
        :param **kwargs : dict, optional
            Additional keyword arguments passed to the `predict` method of the model. Only works with
            univariate target series.
        """
        ts_predicted = super().predict(
            n,
            ts,
            past_covariates,
            future_covariates,
            num_samples,
            verbose,
            predict_likelihood_parameters,
            **kwargs,
        )
        return TimeSeries.from_darts(ts_predicted)



from abc import ABC, abstractmethod
from typing import NoReturn

from ..time_series import TimeSeries


class AbstractBaseProcessor(ABC):
    """Abstract class to define methods to implement
    for a Processor class.
    """

    @abstractmethod
    def process(self, ts: TimeSeries) -> NoReturn:
        """Process time series"""
        raise NotImplementedError



from darts.models import CatBoostModel
from darts.utils.statistics import check_seasonality
import ontime as on


class GenericDetector:
    """
    Generic detector for univariate time series using CatBoost model
    """

    def __init__(self):
        self.model = None
        self.low_quantile = 0.2
        self.high_quantile = 0.8

    def fit(self, ts):
        """
        Fit the model to the given time series as well as a detector
        :param ts: TimeSeries
        :return: self
        """
        # Get information about the time series
        has_seasonality, seasonality = check_seasonality(ts)
        lags = 12 if seasonality == 0 else seasonality

        # Create model
        self.model = on.Model(
            CatBoostModel,
            lags=int(lags),
        )
        self.model.fit(ts)

        # Create detector
        self.detector = on.detectors.quantile(
            low_quantile=self.low_quantile, high_quantile=self.high_quantile
        )
        self.detector.fit(ts)

        return self

    def detect(self, ts):
        """
        Detect anomalies in the given time series
        :param ts: TimeSeries
        :return: TimeSeries
        """
        return self.detector.detect(ts)

    def predetect(self, n):
        """
        Predict n steps into the future and detect anomalies
        :param n: Int number of steps to predict
        :return: TimeSeries
        """
        pred = self.model.predict(n)
        return self.detector.detect(pred)



from typing import Optional, Union

import numpy as np
import pandas as pd
from darts.utils.timeseries_generation import linear_timeseries

from ...abstract import AbstractBaseGenerator
from ...time_series import TimeSeries


class Linear(AbstractBaseGenerator):
    """
    Wrapper around Darts linear time series generator.
    """

    def __init__(self):
        super().__init__()

    def generate(
        self,
        start_value: float = 0,
        end_value: float = 1,
        start: Optional[Union[pd.Timestamp, int]] = pd.Timestamp("2000-01-01"),
        end: Optional[Union[pd.Timestamp, int]] = None,
        length: Optional[int] = None,
        freq: Union[str, int] = None,
        column_name: Optional[str] = "linear",
        dtype: np.dtype = np.float64,
    ):
        """
        Generates a linear time series.

        :param start_value: float
        :param end_value: float
        :param start: Pandas Timestamp or int
        :param end: Pandas Timestamp or int
        :param length: int
        :param freq: str or int
        :param column_name: str
        :param dtype: np.float64
        :return: TimeSeries
        """
        ts = linear_timeseries(
            start_value, end_value, start, end, length, freq, column_name, dtype
        )
        return TimeSeries.from_darts(ts)



from typing import Optional, Tuple, Union, List

from darts.models.forecasting.tcn_model import TCNModel
import pytorch_lightning as pl

from ...abstract import AbstractBaseModel
from ...time_series import TimeSeries


class TCN(TCNModel, AbstractBaseModel):
    """
    Wrapper around Darts TCN model.
    """

    def __init__(
        self,
        input_chunk_length: int,
        output_chunk_length: int,
        kernel_size: int = 3,
        num_filters: int = 3,
        num_layers: Optional[int] = None,
        dilation_base: int = 2,
        weight_norm: bool = False,
        dropout: float = 0.2,
        **kwargs
    ):
        """Temporal Convolutional Network Model (TCN).

        This is an implementation of a dilated TCN used for forecasting, inspired from https://arxiv.org/abs/1803.01271.

        This model supports past covariates (known for `input_chunk_length` points before prediction time).

        Some other arguments can be added to the model by passing them as keyword arguments. To know them, check
        the Darts documentation for the :class:`TCNModel` class.

        input_chunk_length
            Number of past time steps that are fed to the forecasting module.
        output_chunk_length
            Number of time steps the torch module will predict into the future at once.
        kernel_size
            The size of every kernel in a convolutional layer.
        num_filters
            The number of filters in a convolutional layer of the TCN.
        weight_norm
            Boolean value indicating whether to use weight normalization.
        dilation_base
            The base of the exponent that will determine the dilation on every level.
        num_layers
            The number of convolutional layers.
        dropout
            The dropout rate for every convolutional layer. This is compatible with Monte Carlo dropout
            at inference time for model uncertainty estimation (enabled with ``mc_dropout=True`` at
            prediction time).
        **kwargs
            Optional arguments to initialize the pytorch_lightning.Module, pytorch_lightning.Trainer, and
            Darts' :class:`TorchForecastingModel`.

        """
        super().__init__(
            input_chunk_length,
            output_chunk_length,
            kernel_size,
            num_filters,
            num_layers,
            dilation_base,
            weight_norm,
            dropout,
            **kwargs
        )

    def fit(
        self,
        ts: TimeSeries,
        past_covariates: Optional[TimeSeries] = None,
        future_covariates: Optional[TimeSeries] = None,
        val_series: Optional[TimeSeries] = None,
        val_past_covariates: Optional[TimeSeries] = None,
        val_future_covariates: Optional[TimeSeries] = None,
        trainer: Optional[pl.Trainer] = None,
        verbose: Optional[bool] = None,
        epochs: int = 0,
        max_samples_per_ts: Optional[int] = None,
        num_loader_workers: int = 0,
    ):
        """Fit/train the model on one or multiple series.

        This method wraps around :func:`fit_from_dataset()`, constructing a default training
        dataset for this model. If you need more control on how the series are sliced for training, consider
        calling :func:`fit_from_dataset()` with a custom :class:`darts.utils.data.TrainingDataset`.

        Training is performed with a PyTorch Lightning Trainer. It uses a default Trainer object from presets and
        ``pl_trainer_kwargs`` used at model creation. You can also use a custom Trainer with optional parameter
        ``trainer``. For more information on PyTorch Lightning Trainers check out `this link
        <https://pytorch-lightning.readthedocs.io/en/stable/common/trainer.html>`_ .

        This function can be called several times to do some extra training. If ``epochs`` is specified, the model
        will be trained for some (extra) ``epochs`` epochs.

        Below, all possible parameters are documented, but not all models support all parameters. For instance,
        all the :class:`PastCovariatesTorchModel` support only ``past_covariates`` and not ``future_covariates``.
        Darts will complain if you try fitting a model with the wrong covariates argument.

        When handling covariates, Darts will try to use the time axes of the target and the covariates
        to come up with the right time slices. So the covariates can be longer than needed; as long as the time axes
        are correct Darts will handle them correctly. It will also complain if their time span is not sufficient.

        :params ts: TimeSeries
            A series or sequence of series serving as target (i.e. what the model will be trained to forecast)
        :params past_covariates: Optional[TimeSeries]
            Optionally, a series or sequence of series specifying past-observed covariates
        :params future_covariates: Optional[TimeSeries]
            Optionally, a series or sequence of series specifying future-known covariates
        :params val_series: Optional[TimeSeries]
            Optionally, one or a sequence of validation target series, which will be used to compute the validation
            loss throughout training and keep track of the best performing models.
        :params val_past_covariates: Optional[TimeSeries]
            Optionally, the past covariates corresponding to the validation series (must match ``covariates``)
        :params val_future_covariates: Optional[TimeSeries]
            Optionally, the future covariates corresponding to the validation series (must match ``covariates``)
        :params trainer: Optional[pl.Trainer]
            Optionally, a custom PyTorch-Lightning Trainer object to perform training. Using a custom ``trainer`` will
            override Darts' default trainer.
        :params verbose: Optional[bool]
            Optionally, whether to print progress.
        :params epochs: int
            If specified, will train the model for ``epochs`` (additional) epochs, irrespective of what ``n_epochs``
            was provided to the model constructor.
        :params max_samples_per_ts: Optional[int]
            Optionally, a maximum number of samples to use per time series. Models are trained in a supervised fashion
            by constructing slices of (input, output) examples. On long time series, this can result in unnecessarily
            large number of training samples. This parameter upper-bounds the number of training samples per time
            series (taking only the most recent samples in each series). Leaving to None does not apply any
            upper bound.
        :params num_loader_workers: int
            Optionally, an integer specifying the ``num_workers`` to use in PyTorch ``DataLoader`` instances,
            both for the training and validation loaders (if any).
            A larger number of workers can sometimes increase performance, but can also incur extra overheads
            and increase memory usage, as more batches are loaded in parallel.

        :return self:
            Fitted model.
        """
        super().fit(
            ts,
            past_covariates,
            future_covariates,
            val_series,
            val_past_covariates,
            val_future_covariates,
            trainer,
            verbose,
            epochs,
            max_samples_per_ts,
            num_loader_workers,
        )
        return self

    def predict(
        self,
        n: int,
        series: Optional[TimeSeries] = None,
        past_covariates: Optional[TimeSeries] = None,
        future_covariates: Optional[TimeSeries] = None,
        trainer: Optional[pl.Trainer] = None,
        batch_size: Optional[int] = None,
        verbose: Optional[bool] = None,
        n_jobs: int = 1,
        roll_size: Optional[int] = None,
        num_samples: int = 1,
        num_loader_workers: int = 0,
        mc_dropout: bool = False,
        predict_likelihood_parameters: bool = False,
    ) -> TimeSeries:
        """Predict the ``n`` time step following the end of the training series, or of the specified ``series``.

        Prediction is performed with a PyTorch Lightning Trainer. It uses a default Trainer object from presets and
        ``pl_trainer_kwargs`` used at model creation. You can also use a custom Trainer with optional parameter
        ``trainer``. For more information on PyTorch Lightning Trainers check out `this link
        <https://pytorch-lightning.readthedocs.io/en/stable/common/trainer.html>`_ .

        Below, all possible parameters are documented, but not all models support all parameters. For instance,
        all the :class:`PastCovariatesTorchModel` support only ``past_covariates`` and not ``future_covariates``.
        Darts will complain if you try calling :func:`predict()` on a model with the wrong covariates argument.

        Darts will also complain if the provided covariates do not have a sufficient time span.
        In general, not all models require the same covariates' time spans:

        * | Models relying on past covariates require the last ``input_chunk_length`` of the ``past_covariates``
          | points to be known at prediction time. For horizon values ``n > output_chunk_length``, these models
          | require at least the next ``n - output_chunk_length`` future values to be known as well.
        * | Models relying on future covariates require the next ``n`` values to be known.
          | In addition (for :class:`DualCovariatesTorchModel` and :class:`MixedCovariatesTorchModel`), they also
          | require the "historic" values of these future covariates (over the past ``input_chunk_length``).

        When handling covariates, Darts will try to use the time axes of the target and the covariates
        to come up with the right time slices. So the covariates can be longer than needed; as long as the time axes
        are correct Darts will handle them correctly. It will also complain if their time span is not sufficient.

        :param n: int
            The number of time steps after the end of the training time series for which to produce predictions
        :param series: Optional[TimeSeries]
            Optionally, a series or sequence of series, representing the history of the target series whose
            future is to be predicted. If specified, the method returns the forecasts of these
            series. Otherwise, the method returns the forecast of the (single) training series.
        :param past_covariates: Optional[TimeSeries]
            Optionally, the past-observed covariates series needed as inputs for the model.
            They must match the covariates used for training in terms of dimension.
        :param future_covariates: Optional[TimeSeries]
            Optionally, the future-known covariates series needed as inputs for the model.
            They must match the covariates used for training in terms of dimension.
        :param trainer: Optional[pl.Trainer]
            Optionally, a custom PyTorch-Lightning Trainer object to perform prediction. Using a custom ``trainer``
            will override Darts' default trainer.
        :param batch_size: Optional[int]
            Size of batches during prediction. Defaults to the models' training ``batch_size`` value.
        :param verbose: Optional[bool]
            Optionally, whether to print progress.
        :param n_jobs: int
            The number of jobs to run in parallel. ``-1`` means using all processors. Defaults to ``1``.
        :param roll_size: Optional[int]
            For self-consuming predictions, i.e. ``n > output_chunk_length``, determines how many
            outputs of the model are fed back into it at every iteration of feeding the predicted target
            (and optionally future covariates) back into the model. If this parameter is not provided,
            it will be set ``output_chunk_length`` by default.
        :param num_samples: int
            Number of times a prediction is sampled from a probabilistic model. Should be left set to 1
            for deterministic models.
        :param num_loader_workers: int
            Optionally, an integer specifying the ``num_workers`` to use in PyTorch ``DataLoader`` instances,
            for the inference/prediction dataset loaders (if any).
            A larger number of workers can sometimes increase performance, but can also incur extra overheads
            and increase memory usage, as more batches are loaded in parallel.
        :param mc_dropout: bool
            Optionally, enable monte carlo dropout for predictions using neural network based models.
            This allows bayesian approximation by specifying an implicit prior over learned models.
        :param predict_likelihood_parameters: bool
            If set to `True`, the model predict the parameters of its Likelihood parameters instead of the target. Only
            supported for probabilistic models with a likelihood, `num_samples = 1` and `n<=output_chunk_length`.
            Default: ``False``

        :return TimeSeries:
            One time series containing the forecasts of ``series``, or the forecast of the training series
            if ``series`` is not specified and the model has been trained on a single series.
        """
        ts_predicted = super().predict(
            n,
            series,
            past_covariates,
            future_covariates,
            trainer,
            batch_size,
            verbose,
            n_jobs,
            roll_size,
            num_samples,
            num_loader_workers,
            mc_dropout,
            predict_likelihood_parameters,
        )
        return TimeSeries.from_darts(ts_predicted)



from types import SimpleNamespace

COLORS_DICT = {
    "blue_light": "#CADEF7",
    "blue": "#29335C",
    "blue_dark": "#0A2342",
    "green": "#297373",
    "yellow": "#E8C547",
    "red": "#E4572E",
    "red_dark": "#92140C",
    "grey": "#CDD1C4",
}

colors = SimpleNamespace(**COLORS_DICT)



from darts.ad.detectors.threshold_detector import ThresholdDetector

from ...abstract import AbstractBaseDetector
from ...time_series import TimeSeries


class Threshold(ThresholdDetector, AbstractBaseDetector):
    """
    Wrapper around Darts ThresholdDetector.
    """

    def __init__(self, low_threshold=None, high_threshold=None):
        super().__init__(low_threshold, high_threshold)

    def detect(self, ts: TimeSeries) -> TimeSeries:
        """
        Detects anomalies in the given time series.
        :param ts: TimeSeries
        :return: TimeSeries
        """
        ts_detected = super().detect(ts)
        return TimeSeries.from_darts(ts_detected)



""" OnTime API Definition """

# from .abstract import *
from .context import *
from .detectors import detectors

# from .config import *
from .generators import generators

# from .metrics import metrics
from .model import Model
from .models import models

from .plots import plots
from .processors import processors
from .time_series import TimeSeries



from typing import Optional, Union

import numpy as np
import pandas as pd
from darts.utils.timeseries_generation import gaussian_timeseries

from ...abstract import AbstractBaseGenerator
from ...time_series import TimeSeries


class Gaussian(AbstractBaseGenerator):
    """
    Wrapper around Darts gaussian time series generator.
    """

    def __init__(self):
        super().__init__()

    def generate(
        self,
        mean: Union[float, np.ndarray] = 0.0,
        std: Union[float, np.ndarray] = 1.0,
        start: Optional[Union[pd.Timestamp, int]] = pd.Timestamp("2000-01-01"),
        end: Optional[Union[pd.Timestamp, int]] = None,
        length: Optional[int] = None,
        freq: Union[str, int] = None,
        column_name: Optional[str] = "gaussian",
        dtype: np.dtype = np.float64,
    ):
        """
        Generates a gaussian time series.

        :param mean: float or np.ndarray
        :param std: float or np.ndarray
        :param start: Pandas Timestamp or int
        :param end: Pandas Timestamp or int
        :param length: int
        :param freq: str or int
        :param column_name: str
        :param dtype: np.float64
        :return: TimeSeries
        """
        ts = gaussian_timeseries(
            mean, std, start, end, length, freq, column_name, dtype
        )
        return TimeSeries.from_darts(ts)



from .generators import Generators
from .registry.constant import Constant
from .registry.gaussian import Gaussian
from .registry.holiday import Holiday
from .registry.linear import Linear
from .registry.random_walk import RandomWalk
from .registry.sine import Sine

generators = Generators()
generators.load("constant", Constant)
generators.load("gaussian", Gaussian)
generators.load("holiday", Holiday)
generators.load("linear", Linear)
generators.load("random_walk", RandomWalk)
generators.load("sine", Sine)



from .processors import Processors
from .registry.filler import Filler
from .registry.mapper import Mapper
from .registry.windower import Windower

processors = Processors()
processors.load("filler", Filler)
processors.load("mapper", Mapper)
processors.load("windower", Windower)



from abc import ABC, abstractmethod

from ..time_series import TimeSeries


class AbstractBaseDetector(ABC):
    """Abstract class to define methods to implement
    for a Detector class.
    """

    @abstractmethod
    def detect(self, ts: TimeSeries) -> TimeSeries:
        """Detect features"""
        raise NotImplementedError



from typing import Optional, Tuple

from darts.models.forecasting.arima import ARIMA

from ...abstract import AbstractBaseModel
from ...time_series import TimeSeries


class ARIMA(ARIMA, AbstractBaseModel):
    """
    Wrapper around Darts ARIMA model.
    """

    def __init__(
        self,
        p: int = 12,
        d: int = 1,
        q: int = 0,
        seasonal_order: Tuple[int, int, int, int] = (0, 0, 0, 0),
        trend: Optional[str] = None,
        random_state: Optional[int] = None,
        add_encoders: Optional[dict] = None,
    ):
        """
        ARIMA

        ARIMA-type models extensible with exogenous variables (future covariates)
        and seasonal components.

        :param p: int
            Order (number of time lags) of the autoregressive model (AR).
        :param d: int
            The order of differentiation; i.e., the number of times the data
            have had past values subtracted (I).
        :param q : int
            The size of the moving average window (MA).
        :param seasonal_order: Tuple[int, int, int, int]
            The (P,D,Q,s) order of the seasonal component for the AR parameters,
            differences, MA parameters and periodicity.
        :param trend: str
            Parameter controlling the deterministic trend. 'n' indicates no trend,
            'c' a constant term, 't' linear trend in time, and 'ct' includes both.
            Default is 'c' for models without integration, and no trend for models with integration.
        :param add_encoders
            A large number of future covariates can be automatically generated with `add_encoders`.
            This can be done by adding multiple pre-defined index encoders and/or custom user-made functions that
            will be used as index encoders. Additionally, a transformer such as Darts' :class:`Scaler` can be added to
            transform the generated covariates. This happens all under one hood and only needs to be specified at
            model creation.
            Read :meth:`SequentialEncoder <darts.dataprocessing.encoders.SequentialEncoder>` to find out more about
            ``add_encoders``. Default: ``None``. An example showing some of ``add_encoders`` features:

                add_encoders={
                    'cyclic': {'future': ['month']},
                    'datetime_attribute': {'future': ['hour', 'dayofweek']},
                    'position': {'future': ['relative']},
                    'custom': {'future': [lambda idx: (idx.year - 1950) / 50]},
                    'transformer': Scaler()
                }
        """
        super().__init__(p, d, q, seasonal_order, trend, random_state, add_encoders)

    def fit(
        self, ts: TimeSeries, future_covariates: Optional[TimeSeries] = None
    ) -> ARIMA:
        """Fit/train the model on the (single) provided series.

        Optionally, a future covariates series can be provided as well.

        :param ts: TimeSeries
            The model will be trained to forecast this time series. Can be multivariate if the model supports it.
        :param future_covariates: TimeSeries
            A time series of future-known covariates. This time series will not be forecasted, but can be used by
            some models as an input. It must contain at least the same time steps/indices as the target `series`.
            If it is longer than necessary, it will be automatically trimmed.
        :return: self
        """
        super().fit(ts, future_covariates)
        return self

    def predict(
        self,
        n: int,
        ts: Optional[TimeSeries] = None,
        future_covariates: Optional[TimeSeries] = None,
        num_samples: int = 1,
        **kwargs,
    ) -> TimeSeries:
        """If the `series` parameter is not set, forecasts values for `n` time steps after the end of the training
        series. If some future covariates were specified during the training, they must also be specified here.

        If the `series` parameter is set, forecasts values for `n` time steps after the end of the new target
        series. If some future covariates were specified during the training, they must also be specified here.

        :param n: int
            Forecast horizon - the number of time steps after the end of the series for which to produce predictions.
        :param ts: TimeSeries
            Optionally, a new target series whose future values will be predicted. Defaults to `None`, meaning that the
            model will forecast the future value of the training series.
        :param future_covariates: TimeSeries
            The time series of future-known covariates which can be fed as input to the model. It must correspond to
            the covariate time series that has been used with the :func:`fit()` method for training.

            If `series` is not set, it must contain at least the next `n` time steps/indices after the end of the
            training target series. If `series` is set, it must contain at least the time steps/indices corresponding
            to the new target series (historic future covariates), plus the next `n` time steps/indices after the end.
        :param num_samples: int
            Number of times a prediction is sampled from a probabilistic model. Should be left set to 1
            for deterministic models.

        :return: TimeSeries
        """
        ts_predicted = super().predict(n, ts, future_covariates, num_samples, **kwargs)
        return TimeSeries.from_darts(ts_predicted)



from .abstract_base_detector import AbstractBaseDetector
from .abstract_base_generator import AbstractBaseGenerator
from .abstract_base_model import AbstractBaseModel
from .abstract_base_processor import AbstractBaseProcessor



from .models import Models
from .registry.arima import ARIMA
from .registry.catboost import CatBoost
from .registry.tcn import TCN

models = Models()
models.load("arima", ARIMA)
models.load("catboost", CatBoost)
models.load("TCN", TCN)



from typing import Union, List, Optional

from darts.dataprocessing.transformers.window_transformer import (
    WindowTransformer as DartsWindowTransformer,
)

from ...abstract import AbstractBaseProcessor
from ...time_series import TimeSeries


class Windower(AbstractBaseProcessor):
    """
    Wrapper around Darts WindowTransformer.
    https://unit8co.github.io/darts/generated_api/darts.dataprocessing.transformers.window_transformer.html#window-transformer
    """

    def __init__(
        self,
        transforms: Union[dict, List[dict]],
        treat_na: Optional[Union[str, Union[int, float]]] = None,
        forecasting_safe: Optional[bool] = True,
        keep_non_transformed: Optional[bool] = False,
        include_current: Optional[bool] = True,
        name: str = "Windower",
        n_jobs: int = 1,
        verbose: bool = False,
    ):
        """
        A transformer that applies window transformation to a TimeSeries or a Sequence of TimeSeries. It expects a
        dictionary or a list of dictionaries specifying the window transformation(s) to be applied. All series in the
        sequence will be transformed with the same transformations.

        :param transforms: dict or list of dict
            Each dictionary specifies a different window transform. Check the documentation of Darts.
        :param treat_na:
            Specifies how to treat missing values that were added by the window transformations
            at the beginning of the resulting TimeSeries. By default, Darts will leave NaNs in the resulting TimeSeries.
            This parameter can be one of the following: 'dropna', 'backfill', an integer or a float.
        :param forecasting_safe:
            If True, Darts enforces that the resulting TimeSeries is safe to be used in forecasting models as target
            or as feature. The window transformation will not allow future values to be included in the computations
            at their corresponding current timestep. Default is ``True``.
            "ewm" and "expanding" modes are forecasting safe by default.
            "rolling" mode is forecasting safe if ``"center": False`` is guaranteed.
        :param keep_non_transformed:
            ``False`` to return the transformed components only, ``True`` to return all original components along
            the transformed ones. Default is ``False``.
        :param include_current:
            ``True`` to include the current time step in the window, ``False`` to exclude it. Default is ``True``.
        :param name:
            A specific name for the transformer.
        :param n_jobs:
            The number of jobs to run in parallel. Parallel jobs are created only when a ``Sequence[TimeSeries]`` is
            passed as input to a method, parallelising operations regarding different ``TimeSeries``. Defaults to `1`.
        :param verbose:
            Whether to print operations progress.
        """
        super().__init__()
        self._processor = DartsWindowTransformer(
            transforms,
            treat_na=treat_na,
            forecasting_safe=forecasting_safe,
            keep_non_transformed=keep_non_transformed,
            include_current=include_current,
            name=name,
            n_jobs=n_jobs,
            verbose=verbose,
        )

    def process(self, ts: TimeSeries) -> TimeSeries:
        """
        Process the time series.

        :param ts: TimeSeries
        :return: TimeSeries
        """
        return self._processor.transform(ts)



from typing import Optional, Union

import numpy as np
import pandas as pd
from darts.utils.timeseries_generation import sine_timeseries

from ...abstract import AbstractBaseGenerator
from ...time_series import TimeSeries


class Sine(AbstractBaseGenerator):
    """
    Wrapper around Darts sine time series generator.
    """

    def __init__(self):
        super().__init__()

    def generate(
        self,
        value_frequency: float = 0.1,
        value_amplitude: float = 1.0,
        value_phase: float = 0.0,
        value_y_offset: float = 0.0,
        start: Optional[Union[pd.Timestamp, int]] = pd.Timestamp("2000-01-01"),
        end: Optional[Union[pd.Timestamp, int]] = None,
        length: Optional[int] = None,
        freq: Union[str, int] = None,
        column_name: Optional[str] = "sine",
        dtype: np.dtype = np.float64,
    ):
        """
        Generates a sine time series.

        :param value_frequency: float
        :param value_amplitude: float
        :param value_phase: float
        :param value_y_offset: float
        :param start: Pandas Timestamp or int
        :param end: Pandas Timestamp or int
        :param length: int
        :param freq: str or int
        :param column_name: str
        :param dtype: np.float64
        :return: TimeSeries
        """
        ts = sine_timeseries(
            value_frequency,
            value_amplitude,
            value_phase,
            value_y_offset,
            start,
            end,
            length,
            freq,
            column_name,
            dtype,
        )
        return TimeSeries.from_darts(ts)



from darts.dataprocessing.transformers.mappers import Mapper as DartsMapper
from darts.dataprocessing.transformers.mappers import (
    InvertibleMapper as DartsInvertibleMapper,
)

from ...abstract import AbstractBaseProcessor
from ...time_series import TimeSeries


class Mapper(AbstractBaseProcessor):
    """
    Wrapper around Darts Mapper
    https://unit8co.github.io/darts/generated_api/darts.dataprocessing.transformers.mappers.html
    """

    def __init__(self, fn, inverse_fn=None, name="Map", n_jobs=1, verbose=False):
        """Data transformer to apply a custom function and its inverse to a ``TimeSeries``
        (similar to calling :func:`TimeSeries.map()` on each series).

        :param fn:
            Either a function which takes a value and returns a value ie. `f(x) = y`
            Or a function which takes a value and its timestamp and returns a value ie. `f(timestamp, x) = y`.
        :param inverse_fn:
            Similarly to `fn`, either a function which takes a value and returns a value ie. `f(x) = y`
            Or a function which takes a value and its timestamp and returns a value ie. `f(timestamp, x) = y`.
            `inverse_fn` should be such that ``inverse_fn(fn(x)) == x``.        :param name:
        :param name:
            A specific name for the transformer.
        :param n_jobs:
            The number of jobs to run in parallel. Parallel jobs are created only when a `Sequence[TimeSeries]` is
            passed as input to a method, parallelising operations regarding different `TimeSeries`. Defaults to `1`
            (sequential). Setting the parameter to `-1` means using all the available processors.
            Note: for a small amount of data, the parallelisation overhead could end up increasing the total
            required amount of time.
        :param verbose:
            Optionally, whether to print operations progress
        """
        super().__init__()
        self.func = fn

        if inverse_fn is None:
            self._processor = DartsMapper(fn, name=name, n_jobs=n_jobs, verbose=verbose)
        else:
            self._processor = DartsInvertibleMapper(
                fn, inverse_fn, name=name, n_jobs=n_jobs, verbose=verbose
            )

    def process(self, ts) -> TimeSeries:
        """Process the time series.

        :param ts: TimeSeries
        :return: TimeSeries
        """
        return self._processor.transform(ts)

    def inverse_process(self, ts) -> TimeSeries:
        """Inverse process the time series.

        :param ts: TimeSeries
        :return: TimeSeries
        """
        if isinstance(self._processor, DartsInvertibleMapper):
            return self._processor.inverse_transform(ts)
        else:
            raise NotImplementedError(
                "This mapper is not invertible, add a inverse_func to the constructor"
            )



from .detectors import Detectors
from .registry.threshold import Threshold
from .registry.quantile import Quantile

detectors = Detectors()
detectors.load("threshold", Threshold)
detectors.load("quantile", Quantile)



from __future__ import annotations
from abc import ABC, abstractmethod
from typing import Any, NoReturn

from ..time_series import TimeSeries


class AbstractBaseModel(ABC):
    """Abstract class to define methods to implement
    for a Model class inspired by Scikit Learn API.
    """

    def __init__(self):
        pass

    @abstractmethod
    def fit(self, ts: TimeSeries) -> NoReturn:
        """Fit a model"""
        pass

    @abstractmethod
    def predict(self, horizon: Any) -> Any:
        """Usage of the model to predict values"""
        pass



from __future__ import annotations
from typing import List

from darts import TimeSeries as DartsTimeSeries
import pandas as pd
import xarray as xr

from ..plots import plots


class TimeSeries(DartsTimeSeries):
    def __init__(self, xa: xr.DataArray):
        super().__init__(xa)

    def split_by_period(self, period: str) -> List[TimeSeries]:
        """
        Split a time series given a period

        The period is defined with offset aliases from Pandas https://pandas.pydata.org/pandas-docs/stable/user_guide/timeseries.html#offset-aliases

        :param period: period to split by
        :return: List of TimeSeries
        """
        df = self.pd_dataframe()
        splits_df = [g for n, g in df.groupby(pd.Grouper(freq=period))]
        splits_ts = list(map(self.from_dataframe, splits_df))
        return splits_ts

    @staticmethod
    def group_splits(split_ts: List[TimeSeries]) -> TimeSeries:
        """
        Group a list of time series into a single time series

        :param split_ts:  List of TimeSeries
        :return: TimeSeries
        """
        ts = split_ts[0]
        for ts_i in split_ts[1:]:
            ts = ts.append(ts_i)
        return ts

    @staticmethod
    def from_darts(ts):
        """
        Convert a Darts TimeSeries to an OnTime TimeSeries

        :param ts: Darts TimeSeries
        :return: OnTime TimeSeries
        """
        return TimeSeries(ts.data_array())



from ontime.time_series import TimeSeries
from enum import Enum
import pandas as pd


class Profiler:
    """
    This class should not be instantiated.
    This class is used to make a profile from a time series.
    """

    # Define const for column's names
    VALUE_COL = "value"
    TIME_COL = "time"

    # Define the all aggregations possible
    class Aggregation(Enum):
        MEAN = "mean"
        MEDIAN = "median"
        SUM = "sum"

    # Define the all periods possible
    # The first element is the offset alias for split_by_period from ontime.time_series
    # The second element is the format to make the aggregation (**Also users' format**)
    # The third element is the format to convert data to match with TimeSeries format and the chosen period
    class Period(Enum):
        DAILY = ["D", "%H:%M", "%H:%M"]
        WEEKLY = ["W", "%A %H:%M", "%d %H:%M"]
        MONTHLY = ["M", "%d %H:%M", "%d %H:%M"]
        YEARLY = ["Y", "%m%d %H:%M", "%m%d %H:%M"]

    @staticmethod
    def profile(ts: TimeSeries, period: Period, aggregation: Aggregation):
        """
        Make a profile from a time series
        It resumes the time series by a period and an aggregation.
        For example, if you have a time series with a value for each minute over a year,
        you can make a profile by period (day, week, month or year) with an aggregation (mean, median or sum).
        It returns a time series with the new values that are the aggregation of the values that appear at
        the same period (hour:minute, day of the week hour:minute, day hour:minute, month day hour:minute).
        The index is the **first occurrence** of the period.

        :param ts: TimeSeries
        :param period: Period
        :param aggregation: Aggregation
        :return: TimeSeries
        """
        # Define the column names
        _time_col = Profiler.TIME_COL
        _value_col = Profiler.VALUE_COL
        time_agg = "time_formatted"

        # Split the time series by week
        split_ts = ts.split_by_period(period.value[0])

        # Create a list of DataFrames from the split time series
        data_frames = []
        for time_series in split_ts:
            df = time_series.pd_dataframe()
            col = df.columns[0]
            value = df[col].values
            temp_df = pd.DataFrame(
                {
                    time_agg: df.index.strftime(period.value[1]),
                    _time_col: df.index,
                    _value_col: value,
                }
            )
            data_frames.append(temp_df)

        # Concatenate all the DataFrames in the list into one DataFrame
        data = pd.concat(data_frames)

        # Group the DataFrame by the day of the week and calculate the mean
        grouped = data.groupby([time_agg]).agg(
            {_value_col: aggregation.value, _time_col: "first"}
        )

        # Reset the index, making 'period' the new index column
        grouped.reset_index(inplace=True)

        # Convert the 'period' column to datetime using the custom format
        grouped[_time_col] = pd.to_datetime(grouped[_time_col], format=period.value[2])

        # Return a TimeSeries from the grouped DataFrame
        return TimeSeries.from_dataframe(
            grouped, time_col=_time_col, value_cols=_value_col
        )



from typing import Optional, Union

import numpy as np
import pandas as pd
from darts.utils.timeseries_generation import holidays_timeseries

from ...abstract import AbstractBaseGenerator
from ...time_series import TimeSeries


class Holiday(AbstractBaseGenerator):
    """
    Wrapper around Darts holiday time series generator.
    """

    def __init__(self):
        super().__init__()

    def generate(
        self,
        time_index: pd.DatetimeIndex,
        country_code: str,
        prov: str = None,
        state: str = None,
        column_name: Optional[str] = "holidays",
        until: Optional[Union[int, str, pd.Timestamp]] = None,
        add_length: int = 0,
        dtype: np.dtype = np.float64,
    ):
        """
        Generates a holiday time series.

        :param time_index: Pandas DatetimeIndex
        :param country_code: str
        :param prov: str
        :param state: str
        :param column_name: str
        :param until: int, str or Pandas Timestamp
        :param add_length: int
        :param dtype: np.float64
        :return: TimeSeries
        """
        ts = holidays_timeseries(
            time_index, country_code, prov, state, column_name, until, add_length, dtype
        )
        return TimeSeries.from_darts(ts)



from ..utils import Registry


class DynamicClass(Registry):
    def __init__(self):
        super().__init__()

    def load(self, name, my_class):
        """
        Load a class dynamically
        :param name: str
        :param my_class: the class to load
        """
        self.register(name, my_class)
        self.__setattr__(name, my_class)

    def load_registry(self):
        """
        Load all classes stored in the registry
        """
        for name, my_class in self.registry.items():
            self.__setattr__(name, my_class)
